#C+
# Copyright 1997 by Softway Systems, Inc.  All rights reserved.
# 
# This Software Source is the unpublished, valuable, confidential property
# of Softway Systems, Inc.  Use is authorised only under the terms and
# conditions set out in the source code licence agreement protecting this
# Software Source.  Any unauthorised use or disclosure of this Software 
# Source is prohibited.
#
# THIS SOFTWARE IS PROVIDED ``AS IS'' BY SOFTWAY SYSTEMS, INC. AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL SOFTWAY SYSTEMS, INC. BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#C-
#
# makefile 
#
# Created by Doug Moen, Jan 97
# Updated for "clean" gnu tools.  Donn, March 97.
# Undated several more times, March-July 97 to add additional subdirs.

# The objective of this script is to provide a bridge between 
# the GNU build conventions and ours.  Given the complexity of
# the GNU scripts (and a desire to be able to generate them using
# the GNU tools as things change), the actual Makefiles are not changed
# at all.

BINUTILS_PROGS=nm ar strings strip size
BINUTILS_SDK_PROGS=objdump
GAS_PROGS=as gasp
LD_PROGS=ld
BYACC_PROGS=byacc
GCC_PROGS=this_isn't_right_yet gcc cc1 cpp protoize gcov libgcc.a
GDB_PROGS=gdb
BUILD_CFLAGS=-O2

targets:
	@( echo " ";\
	echo "Targets available:";\
	echo " ";\
	echo "  build   - build everything locally";\
	echo "	clean   - remove all intermediate targets.";\
	echo "	clobber - remove all targets.";\
	echo "DO NOT USE install - Install basic utilities in $(DESTDIR)";\
	echo "DO NOT USE install_sdk - Install sdk product in $(DESTDIR)";\
	echo "	run_configure - Run GNU configure on each directory";\
	echo " ";\
	)

# When we start a build, we touch some stamp files to keep the generated
# Makefiles from rebuilding some stuff we don't want rebuilt.
# (We don't want to rerun configure for production builds.)

build: build_libiberty build_bfd build_opcodes build_binutils build_byacc  \
       build_gas build_ld build_gcc build_readline build_gdb \
       build_librx build_libstdc++ build_libio build_libg++ build_libexec

# mmalloc isn't quite ready for prime time yet.

build_libiberty:
	cd libiberty; make CFLAGS="$(BUILD_CFLAGS)"

build_byacc:
	cd byacc; make CFLAGS="$(BUILD_CFLAGS)"

build_gas:
	cd gas; make CFLAGS="$(BUILD_CFLAGS)"
	mkdir -p local_bin
	cp gas/as.new local_bin/as

build_ld:
	if [ ! -f ld/stamp-h ]; then touch ld/stamp-h; fi
	cd ld; make CFLAGS="$(BUILD_CFLAGS)"
	mkdir -p local_bin
	cp ld/ld.new local_bin/ld

# g77 cannot be built by this rule, because it's used with both MSVC
# and gcc as compilers, and g77 can only (as of 0.5.21) be built with
# gcc.

# after putting the commands where we can find them, we also salt
# away a few headers.

build_gcc:
	cd gcc; make CFLAGS="$(BUILD_CFLAGS)" PATH=$(PWD)/local_bin:$(PATH) \
		     LANGUAGES="c objective-c proto gcov c++"
	cp gcc/xgcc local_bin/gcc
	cp gcc/cc1 local_bin/cc1
	cp gcc/cc1plus local_bin/cc1plus
	cp gcc/libgcc.a local_bin/libgcc.a
	cp gcc/cpp local_bin/cpp
	mkdir -p local_inc
	# These are needed by libg++
	cp -f gcc/include/new local_inc
	cp -f gcc/include/new.h local_inc
	cp -f gcc/include/exception local_inc


# Build g77 (only).

build_g77:
	cd gcc; make CFLAGS="$(BUILD_CFLAGS)" PATH=$(PWD)/local_bin:$(PATH) \
		     LANGUAGES="f77" CC="$(CC)"
	cd asa; make asa CC="$(CC)"

build_readline:
	cd readline; make CFLAGS="$(BUILD_CFLAGS)"

# gdb's self-test assumes it's debuggable.
# Also... configure of gdb needs tcl and tk, or it won't configure
# them in.
build_gdb: build_readline build_librx build_tcl build_tk
	cd gdb; make CFLAGS="$(BUILD_CFLAGS) -g"
	mkdir -p local_bin
	cp gdb/gdb local_bin/gdb

build_bfd:
	if [ ! -f bfd/stamp-h ]; then touch binutils/stamp-h; fi
	cd bfd; make CFLAGS="$(BUILD_CFLAGS)"

build_opcodes:
	if [ ! -f opcodes/stamp-h ]; then touch opcodes/stamp-h; fi
	cd opcodes; make CFLAGS="$(BUILD_CFLAGS)"

build_binutils:
	if [ ! -f binutils/stamp-h ]; then touch binutils/stamp-h; fi
	cd binutils; make CFLAGS="$(BUILD_CFLAGS)"

build_librx:
	cd librx; make CFLAGS="$(BUILD_CFLAGS)"

build_libio:
	cd libio; make CFLAGS="$(BUILD_CFLAGS)" PATH=$(PWD)/local_bin:$(PATH)

build_libg++:
	cd libg++; make CFLAGS="$(BUILD_CFLAGS)" \
	    PATH=$(PWD)/local_bin:$(PATH) \
	    CXXFLAGS="$(BUILD_CFLAGS) -I$(PWD)/local_inc"

build_libstdc++:
	cd libstdc++; make CFLAGS="$(BUILD_CFLAGS)" PATH=$(PWD)/local_bin:$(PATH)

build_tcl:
	cd tcl; make CFLAGS="$(BUILD_CFLAGS)"

build_tk: build_tcl
	cd tk; make CFLAGS="$(BUILD_CFLAGS)"

build_libexec:
	cd libexec; make CFLAGS="$(BUILD_CFLAGS)"

build_tools: build_tcl
	cd expect; make CFLAGS="$(BUILD_CFLAGS)"
	cd dejagnu; make CFLAGS="$(BUILD_CFLAGS)"

total_build: build build_tools
	cd texinfo; make CFLAGS="$(BUILD_CFLAGS)"

# The clean/clobber targets translate to the GNU conventions
clean:
	cd binutils; make mostlyclean
	cd bfd; make mostlyclean
	cd opcodes; make mostlyclean
	cd libiberty; make mostlyclean
	cd byacc; make mostlyclean
	cd gas; make mostlyclean
	cd gcc; make mostlyclean
	cd ld; make mostlyclean
	cd tcl; make mostlyclean
	cd tk; make mostlyclean
	cd expect; make mostlyclean
	cd texinfo; make mostlyclean
	cd readline; make mostlyclean
	cd gdb; make mostlyclean
	cd librx; make mostlyclean
	cd libio; make mostlyclean
	cd libexec; make mostlyclean
	cd libg++; make mostlyclean
	cd libstdc++; make mostlyclean

clobber:
	cd binutils; make clean
	cd bfd; make clean
	cd opcodes; make clean
	cd libiberty; make clean
	cd byacc; make clean
	cd gas; make clean
	cd gcc; make clean
	cd ld; make clean
	cd tcl; make clean
	cd tk; make clean
	cd expect; make clean
	cd texinfo; make clean
	cd readline; make clean
	cd gdb; make clean
	cd librx; make clean
	cd libio; make clean
	cd libexec; make clean
	cd libstdc++; make clean
	cd libg++; make clean

install: install_binutils

install_sdk: install install_gas install_ld install_byacc install_binutils_sdk \
        install_gcc

install_binutils:
	cd binutils;\
	cp nm.new nm;\
	cp strip.new strip;\
	for i in $(BINUTILS_PROGS); do\
	    $(INSTALL) $$i $(DESTDIR); \
	done

install_binutils_sdk:
	cd binutils;\
	for i in $(BINUTILS_SDK_PROGS); do\
	    $(INSTALL) $$i $(DESTDIR); \
	done

install_gas:
	cd gas;\
	cp as.new as;\
	cp gasp.new gasp;\
	for i in $(GAS_PROGS); do\
	    $(INSTALL) $$i $(DESTDIR); \
	done

install_ld:
	cd ld;\
	cp ld.new ld;\
	for i in $(LD_PROGS); do\
	    $(INSTALL) $$i $(DESTDIR); \
	done

install_byacc:
	cd byacc;\
	for i in $(BYACC_PROGS); do\
	    $(INSTALL) $$i $(DESTDIR); \
	done

install_gcc:
	cd gcc;\
	for i in $(GCC_PROGS); do \
	    $(INSTALL) $$i $(DESTDIR); \
	done

install_gdb:
	cd gdb;\
	for i in $(GDB_PROGS); do \
	    $(INSTALL) $$i $(DESTDIR); \
	done

install_librx:
	dont know how to do that yet.

fullbuild: clobber build install

ARCH=
#i386 or alpha

CONFIGURE=CC="cc -D__STDC__" \
          CONFIG_SHELL=$$INTERIX_ROOT/bin/sh \
	  RANLIB=":"\
          ./configure  --verbose --host=$(ARCH)-pc-interix \
          --disable-shared \
          --target=$(ARCH)-pc-interix  \
          --enable-targets=i386-pc-interix,alpha-pc-interix 

gcc_configure:
	if [ $(ARCH)x = x ]; then echo MUST SET ARCH; exit 1; fi
	cd gcc; rm -f config.cache; $(CONFIGURE)

gdb_configure:
	cd gdb; rm -f config.cache; $(CONFIGURE) # --disable-gdbtk

libexec_configure:
	cd libexec; rm -f config.cache; $(CONFIGURE)

run_configure: gcc_configure gdb_configure libexec_configure
	cd binutils; rm -f config.cache; $(CONFIGURE)
	cd bfd; rm -f config.cache; $(CONFIGURE)
	cd bfd; make headers
	cd opcodes; rm -f config.cache; $(CONFIGURE)
	cd libiberty; rm -f config.cache; $(CONFIGURE)
	cd byacc; rm -f config.cache; $(CONFIGURE)
	cd gas; rm -f config.cache; $(CONFIGURE)
	cd ld; rm -f config.cache; $(CONFIGURE)
	cd readline; rm -f config.cache; $(CONFIGURE)
	cd librx; rm -f config.cache; $(CONFIGURE)
	cd libio; rm -f config.cache; $(CONFIGURE)
	cd libg++; rm -f config.cache; $(CONFIGURE)
	cd libstdc++; rm -f config.cache; $(CONFIGURE)
	cd tcl; rm -f config.cache; CFLAGS="-O2 -g" $(CONFIGURE)
	cd tk; rm -f config.cache; $(CONFIGURE)
	cd asa; $(CONFIGURE)
	
total_configure: run_configure
	cd expect; rm -f config.cache; $(CONFIGURE)
	cd dejagnu; rm -f config.cache; $(CONFIGURE)
	cd texinfo; rm -f config.cache; $(CONFIGURE)

# NOTE:
#   Fullbuild may need to make sure that *everything* gets built; specifically
#   in bfd, does "make headers" (which does things in the DOC directory),
#   do the right thing?
