/* Machine-dependent dynamic relocation inline functions.  Alpha version.
   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#ifndef dl_machine_h
#define dl_machine_h

/*                      
 * Calculate how much off the link address we are loaded.
 * We want to keep relocateable expressions out of the text
 * segment, so go thru the GOT.
 */                             
static inline VMA __attribute__ ((unused))
machine_load_offset (void)     
{
  /* This code is written to work "right" only BEFORE it's relocated, and
     should be called once (and the result retained) before self-relocation
     starts. */

  VMA addr;

  asm("   br %0,1f\n"
      "1: lda $28,1b\n" 
      "   subl %0,$28,%0\n"
      : "=r"(addr) : : "$28");
  return addr;
}

static inline void
no_relocate_message(VMA load_offs)
{
  /* This is intended to be VERY nonportable code that will run when
     the code has not been relocated and has not been loaded at its
     default address. */

  extern VMA __imp_write;
  extern VMA __imp__exit;
  static char msg [] = 
    "ld.so is not at its default address and cannot be relocated\n";
  char *pmsg;
  void (*pwrite)(int, const char *, int);
  void (*pexit)(int);

  pmsg = msg + load_offs;
  pwrite = (void *)*(VMA *)((VMA)&__imp_write + load_offs);
  (*pwrite)(2, pmsg, sizeof(msg));
  pexit = (void *)*(VMA *)((VMA)&__imp__exit + load_offs);
  (*pexit)(255);
}

/* Set up the loaded object described by L so its unrelocated PLT
   entries will jump to the on-demand fixup code in dl-runtime.c.  */

static inline void __attribute__ ((unused))
machine_runtime_setup (struct link_map *l, int lazy)
{
  VMA got;
  extern void _dl_runtime_resolve (VMA);

  if (l->l_info[DT_JMPREL] && lazy)
    {
      /* Init the first few words in the GOT to hook to this runtime resolve
	 code */
      got = (VMA) (l->l_addr + l->l_info[DT_PLTGOT]);

      /* This function will be called to perform the relocation.  */
      *(VMA *)(got + 16) = (VMA) &_dl_runtime_resolve;

      /* Identify this shared object */
      *(VMA *)(got + 24) = (VMA) l;
    }
}

/* This code is used in dl-runtime.c to call the `fixup' function
   and then redirect to the address it returns.  */

#define MACHINE_RUNTIME_TRAMPOLINE asm ( \
"/* Trampoline for _dl_runtime_resolver */
	.globl _dl_runtime_resolve
	.ent _dl_runtime_resolve
_dl_runtime_resolve:
	lda	$sp, -168($sp)
	.frame	$sp, 168, $26, 0
	/* Preserve all registers that C normally doesn't.  */
	stq	$26, 0($sp)
	stq	$0, 8($sp)
	stq	$1, 16($sp)
	stq	$2, 24($sp)
	stq	$3, 32($sp)
	stq	$4, 40($sp)
	stq	$5, 48($sp)
	stq	$6, 56($sp)
	stq	$7, 64($sp)
	stq	$8, 72($sp)
	stq	$16, 80($sp)
	stq	$17, 88($sp)
	stq	$18, 96($sp)
	stq	$19, 104($sp)
	stq	$20, 112($sp)
	stq	$21, 120($sp)
	stq	$22, 128($sp)
	stq	$23, 136($sp)
	stq	$24, 144($sp)
	stq	$25, 152($sp)
	stq	$29, 160($sp)
	.mask	0x27ff01ff, -168
	/* If we set up a GP, it'd be here */
	.prologue 1
	/* Set up the arguments for _dl_runtime_resolve. */
	/* $16 = link_map out of plt0 */
	ldq	$16, 8($27)
	/* $17 = offset of reloc entry */
	mov	$28, $17
	/* Do the fixup */
	bsr	$26, fixup  /* or ..ng if we do GP */
	/* Move the destination address into position.  */
	mov	$0, $27
	/* Restore program registers.  */
	ldq	$26, 0($sp)
	ldq	$0, 8($sp)
	ldq	$1, 16($sp)
	ldq	$2, 24($sp)
	ldq	$3, 32($sp)
	ldq	$4, 40($sp)
	ldq	$5, 48($sp)
	ldq	$6, 56($sp)
	ldq	$7, 64($sp)
	ldq	$8, 72($sp)
	ldq	$16, 80($sp)
	ldq	$17, 88($sp)
	ldq	$18, 96($sp)
	ldq	$19, 104($sp)
	ldq	$20, 112($sp)
	ldq	$21, 120($sp)
	ldq	$22, 128($sp)
	ldq	$23, 136($sp)
	ldq	$24, 144($sp)
	ldq	$25, 152($sp)
	ldq	$29, 160($sp)
	/* Flush the Icache after having modified the .plt code.  */
	imb	/* may not be MP safe... some doc'n indicates that */
	/* Clean up and turn control to the destination */
	lda	$sp, 168($sp)
	jmp	$31, ($27)
	.end _dl_runtime_resolve");

#endif /* dl_machine */

#ifdef RESOLVE

/* Perform the relocation specified by RELOC and SYM (which is fully resolved).
   MAP is the object containing the reloc.  */

extern struct link_map _dl_rtld_map;

/* Fix up the instructions of a PLT entry to invoke the function
   rather than the dynamic linker.  */
static inline void
alpha_fix_plt(struct link_map *l,
		  int plt_index,
		  VMA value)
{
  VMA *plte;
  VMA edisp;

  plte = (VMA *)(l->l_addr + l->l_info[DT_PLTGOT] + 32);

  plte += (plt_index*3);

  /* Find the displacement from the plt entry to the function. 
     (as if we were brancking from the first word). */
  edisp = (VMA)(value - (VMA)&plte[1]) / 4;

  if (edisp >= -0x100000 && edisp < 0x100000)
    {
      /* If we are in range, use br to perfect branch prediction and
	 elide the dependency on the address load.  This case happens,
	 e.g., when a shared library call is resolved to the same library.  */

      /* Emit "br $31,function" */
      plte[1] = 0xc3e00000 | (edisp & 0x1fffff);
    }
  else
    {
      /* Don't bother with the hint since we already know the hint is
	 wrong.  Eliding it prevents the wrong page from getting pulled
	 into the cache.  */

      int hi, lo;
      lo = (short)value;
      hi = (value - lo) >> 16;

      /* Emit "ldah $27,H(0)" */
      plte[0] = 0x277f0000 | (hi & 0xffff);

      /* Emit "lda $27,L($27)" */
      plte[1] = 0x237b0000 | (lo & 0xffff);

      /* Emit "jmp $31,($27)" */
      plte[2] = 0x6bfb0000;
    }

  /* At this point, if we've been doing runtime resolution, Icache is dirty.
     This will be taken care of in _dl_runtime_resolve.  If instead we are
     doing this as part of non-lazy startup relocation, that bit of code
     hasn't made it into Icache yet, so there's nothing to clean up.  */
}

static inline VMA
machine_rel (struct link_map *map,
		 const PE_reloc *reloc, VMA symtab)
{
  VMA *const reloc_addr = (void *) (map->l_addr + reloc->r_vaddr);
  const PE_syment *sym;

  static VMA sym_value;
  static const PE_syment *lastSym;

  static VMA refhi_sym_value;
  static int refhi_valid;

  /* We cannot use a switch here because we cannot locate the switch
     jump table until we've self-relocated.  */

#if 0
  /* This shouldn't happen, but in case of a need for fail-soft, code
     is retained. */
  if (reloc->r_type == R_ALPHA_NONE)
    return 0;
#endif

  /* Relative relocs, first */
  if (reloc->r_type == ALPHA_DYN_R_REL_32)
    {
#ifndef RTLD_BOOTSTRAP
      /* Already done in dynamic linker.  */
      if (map == &_dl_rtld_map)
	  return 0;
#endif
      *reloc_addr += map->l_offs;
      return 0;
    }

  if (reloc->r_type == ALPHA_DYN_R_REL_HIONLY)
    {
#ifndef RTLD_BOOTSTRAP
      /* Already done in dynamic linker.  */
      if (map == &_dl_rtld_map)
	  return 0;
#endif
      *(short *)reloc_addr += (map->l_offs>>16);
      return 0;
    }

  if (reloc->r_type == ALPHA_DYN_R_REL_REFHI)
    {
      /* REFHI needs to prepare for the following PAIR;
	 for a relative relocation, there's nothing to do,
	 because there's no symbol involved. */
      return 0;
    }

  if (reloc->r_type == ALPHA_DYN_R_REL_REFLO)
    {
#ifndef RTLD_BOOTSTRAP
      /* Already done in dynamic linker.  */
      if (map == &_dl_rtld_map)
	  return 0;
#endif
      /* a little-endian shortcut */
      *(short *)reloc_addr += map->l_offs;
      return 0;
    }

  if (reloc->r_type == ALPHA_DYN_R_REL_PAIR)
    {
      VMA val;
#ifndef RTLD_BOOTSTRAP
      /* Already done in dynamic linker.  */
      if (map == &_dl_rtld_map)
	  return 0;
#endif
      val = (*(short *)reloc_addr << 16) + map->l_offs + (short)reloc->r_symndx;
      /* Deal with signed 16 bit arithmetic issues */
      if (val & 0x8000) val += 0x10000;
      val >>= 16;
      *(short *)reloc_addr = val;
      return 0;
    }

  /* Now, symbolic relocs; compute the additional info they need */

  /* Pair really isn't symbolic (that is, we don't look up a symbol
     for it; REFHI handles that.) */
  if (reloc->r_type == IMAGE_REL_ALPHA_PAIR)
    {
      VMA val;
      _dl_assert (refhi_valid);
      refhi_valid = 0;

      val = refhi_sym_value + reloc->r_symndx;
      /* Deal with signed 16 bit arithmetic issues */
      if (val & 0x8000) val += 0x10000;
      val >>= 16;
      *(short *)reloc_addr = val;
      return 0;
    }

  /* Now the real symbolic relocations */

  /* We only use the low 24 bits of the symbol table index; details
     in bfd/coff-nt_alpha.c and COMB reloc below. */
  sym = (PE_syment *)(symtab + 
     (((VMA)(reloc->r_symndx))&0xffffff) * PE_SYMENT_SIZE);
  if (sym != lastSym) 
    {
      const char *pSymName;
      char tmpName[SYMNMLEN+1];
      VMA loadbase;
      VMA undo=0;

      lastSym = sym;

#ifndef RTLD_BOOTSTRAP /* [ */
      if (sym->n_zeroes == 0) 
	{
	  pSymName = strtab + sym->n_offset;
	}
      else 
	{
	  *(PE_nameunion *)(tmpName) = sym->_n;
	  tmpName[SYMNMLEN] = 0;
	  pSymName = tmpName;
	}
#endif /* ] */

#ifndef RTLD_BOOTSTRAP /* [ */
      if (map == &_dl_rtld_map)
	/* Undo the relocation done here during bootstrapping.  Now we will
	   relocate it anew, possibly using a binding found in the user
	   program or a loaded library rather than the dynamic linker's
	   built-in definitions used while loading those libraries.  */
	undo = map->l_offs + sym->n_value;
#endif /* ] */
      loadbase = RESOLVE (&sym,
	  reloc->r_type == ALPHA_DYN_R_JMP_SLOT ? DL_LOOKUP_NOPLT : 0);
      sym_value = (sym ? loadbase + sym->n_value : 0) - undo;
    }

  if (reloc->r_type == ALPHA_R_REFLONG)
    {
      *reloc_addr += sym_value;
      return 0;
    }

  if (reloc->r_type == ALPHA_DYN_R_COMB)
    {
      VMA val;
      long offset;
      VMA *reloc_addr2;
      /* Single combined relocation */

      /* shift right 24 and left 2 */
      offset = reloc->r_symndx >> 22;

      reloc_addr2 = (VMA *)((VMA)reloc_addr + offset);
      val = sym_value + *(short *)reloc_addr2;
      *(short *)reloc_addr2 = val;

      /* Deal with signed 16 bit arithmetic issues */
      if (val & 0x8000) val += 0x10000;
      val >>= 16;
      *(short *)reloc_addr = val;
      return 0;
    }

  if (reloc->r_type == IMAGE_REL_ALPHA_REFHI)
    {
      /* REFHI needs to prepare for the following PAIR;
	 for a relative relocation, there's nothing to do,
	 because there's no symbol involved. */
      _dl_assert (!refhi_valid);
      refhi_valid = 1;
      refhi_sym_value = sym_value;
      return 0;
    }

  if (reloc->r_type == IMAGE_REL_ALPHA_REFLO)
    {
      /* a little-endian shortcut */
      *(short *)reloc_addr += sym_value;
      return 0;
    }


  if (reloc->r_type == ALPHA_DYN_R_JMP_SLOT)
    {
      /* Perform a symbolic reloc on the .got entry */
      /* Be sure to modify machine_rel the same way you modify this */
      alpha_fix_plt (map, reloc->r_vaddr, sym_value);
      return sym_value;
    }

  _dl_assert (! "unexpected dynamic reloc type");
}

/* This is simply performance; we can strip a lot out if we know it's
   a JMP_REL for runtime */

static inline VMA
machine_relplt (struct link_map *map,
		 const PE_reloc *reloc, VMA symtab)
{
  VMA loadbase, sym_value;
  const PE_syment *sym;

  const char *pSymName;
  char tmpName[SYMNMLEN+1];

  sym = (PE_syment *)(symtab + 
     (((VMA)(reloc->r_symndx))&0xffffff) * PE_SYMENT_SIZE);
  if (sym->n_zeroes == 0) {
     pSymName = strtab + sym->n_offset;
  }
  else {
     *(PE_nameunion *)(tmpName) = sym->_n;
     tmpName[SYMNMLEN] = 0;
     pSymName = tmpName;
  }

  _dl_assert (reloc->r_type == ALPHA_DYN_R_JMP_SLOT);

  loadbase = RESOLVE (&sym, DL_LOOKUP_NOPLT);
  sym_value = loadbase + sym->n_value;

  /* Perform a symbolic reloc on the .got entry */
  /* Be sure to modify machine_rel the same way you modify this */
  alpha_fix_plt (map, reloc->r_vaddr, sym_value);
  return sym_value;
}

static inline void
machine_lazy_rel (struct link_map *map, const PE_reloc *reloc)
{
  VMA *const reloc_addr = (void *) (map->l_addr + reloc->r_vaddr);

  if (reloc->r_type == ALPHA_DYN_R_JMP_SLOT)
    {
      /* Nothing to do here; it all happens at runtime */
    }
#if 0  /* see above */
  else if (r_info == R_ALPHA_NONE)
    return;
#endif
  else
    _dl_assert (! "unexpected PLT reloc type");
}

#endif /* RESOLVE */
