/* Copyright (C) 2000 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/* A minimal version of crt0 that sets up for a call to _dl_start, and
   handles the almost-recursion caused by this being the entry point for
   ld.so, which means it can both be the "real" entry point and called
   from the real crt0.o.  It relies on the fact that the first two
   arguments to the startup fuction are zero when called by "real"
   startup, and non-zero when called from crt0.o.

   DT. 10/98
*/

#include <unistd.h>
#include <interix/ldsostartup.h>

/* Use strong symbols. */
#define strcmp _strcmp

extern char * _CRTAPI1 __PdxGetCmdLine(void);  

/* These variables are intended to be the ones in the main program (if
   one is called).  Don't use them until you know they are set up,
   and then with caution because they may or may not be what you think
   they are.  */
extern char **environ;
extern int errno;


extern void __stdcall __PdxInitializeData(int *Perrno, char ***Penviron );

void
__minimalPosixProcessStartup(int startedBefore, startupData *incomingStartup)
{
    char **ppch;
    char *loc_cmdlin;  /* Local copy */
    unsigned long ImageBase;
    int loc_errno;
    long userEntry;
    startupData localStartup;

    /* This is the entry point to ld.so.  Whether we were called from
       crt0 or directly by the system is determined by the first argument;
       the system provides a zero.  crt0 provides a 1 (any non-zero). */

    /* We have VERY little to work with at this point; locals and
       functions in libcpsx.a, and a few others, are it.  (NO libc!.) */
       
    if (startedBefore == 0) {

	/* Not from crt0; we have to set up a minimal environment; in
	   particular housekeep the command line arguments.  Note that
	   at this point, the real "main" hasn't been loaded yet, so
	   we don't know where things like errno and environ belong. */
	loc_cmdlin = __PdxGetCmdLine();

	ppch = (char **)loc_cmdlin;
	localStartup.argv = ppch;

	/* normalize argv pointers */
	localStartup.argc = 0;
	while (NULL != *ppch) {
		*ppch += (int)loc_cmdlin;
		++localStartup.argc;
		++ppch;
	}

	/* normalize environ pointers */
	++ppch;
	localStartup.environ = ppch;

	while (NULL != *ppch) {
		*ppch = *ppch + (int)loc_cmdlin;
		++ppch;
	}

	localStartup.pErrno = &loc_errno;

	/* Make a good guess; the 65k page we're on */
	ImageBase = (unsigned long)(&__minimalPosixProcessStartup) & 0xffff0000;

	/* Tell libc where errno and environ live, for now; we'll tell
	   it the truth in crt0 (if we ever get there), but ld.so needs 
	   usable values now. */
	__PdxInitializeData(localStartup.pErrno, &localStartup.environ);

	localStartup.ldsoName = "";
	localStartup.ldsoLoadAddr = ImageBase;

	/* ld.so needs a viable errno while it's starting up; we provide
	   the address of the one we passed to psxdll above.  We'll switch
	   the DLL over to the global errno in _mainCRTStartup().
	   ld.so takes care of itself just before returning. */

	_dl_start(1, &localStartup, &userEntry);

	/* This is the main program entry point that dl_start figured out
	   for us.   The argument of 1 tells crt0 that ld.so has already
	   run, and not to call it again. */
	((long(*)())userEntry)(1, &localStartup);
    }
    else {
	/* From crt0... just pass on the already housekept arguments,
	   and &userEntry to fill in the arg list. */

        _dl_start(0, incomingStartup, &userEntry);

	/* We'll return, and a call to _mainCRTStartup will occur in crt0 */
    }
}
