/* Minimal replacements for basic facilities used in the dynamic linker.
   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <syscall.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>
#include <string.h>
#include <link.h>

/* The __sys_alloc() family of entry points work out of a different memory
   pool than libc's malloc().  Since we need memory before libc is ready,
   we use them.   Don't MIX the two... it won't work.  */

extern void * __sys_alloc(long);
extern void * __sys_free(void *);
extern void * __sys_realloc(void *, long);

void * _dl_malloc(int size) 
{
  void * res;
  res = __sys_alloc(size);
  return res;
}

void _dl_free(void *mem) 
{
  __sys_free(mem);
}

void _dl_realloc(void *mem, int size) 
{
  __sys_realloc(mem, size);
}

/* Local version of `strdup' function.  */
char *
_dl_strdup (const char *s)
{
  size_t len = strlen (s) + 1;
  void *new = _dl_malloc (len);

  if (new == NULL)
    return NULL;

  return (char *) memcpy (new, s, len);
}

/* Because environ isn't set up yet, we have to use our own, using our
   private copy of environ. */
char * getenv(char * symbol)
{
  char * p;
  char * q;
  char **e;

  extern char **_dl_environ;

  e=_dl_environ;

  while ((p = *e++) != NULL) {
    q = symbol;
    while(*p && *q && *p == *q) {q++; p++;}
    if(!*p || *p != '=' || *q) continue;
    return p+1;
  }
  return 0;
}

#ifndef NDEBUG

/* Define our own assert failure function which doesn't use stdio. */
void
__dl_assert (const void *assertion,
	       const void *file, unsigned int line)
{
  char buf[64];
  buf[sizeof buf - 1] = '\0';
  _dl_sysdep_fatal ("BUG IN DYNAMIC LINKER ld.so: ",
		    (char*)file,": ", _itoa (line, buf + sizeof buf - 1, 10, 0),
		    ": ", "Assertion `", (char *)assertion, "' failed!\n",
		    NULL);

}

#endif

/* Per-architecture minimal functions go into the header below;
   search paths are set up to get the right copy. */
#include "dl-minimal.h"
