/* Relocate a shared object and resolve its references to other loaded objects.
   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "syscall.h"
#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>

#include <link.h>
#include "dynamic-link.h"
#include "dl-internal.h"

void
_dl_relocate_object (struct link_map *l, struct link_map *scope[], int lazy)
{
  PE_filehdr *pFileHdr;
  PE_scnhdr  *pSecHdr;
  int i;

  if (l->l_relocated)
    return;

  pFileHdr = (PE_filehdr *)(l->l_addr);
  pSecHdr = (PE_scnhdr *)(l->l_addr + sizeof(PE_filehdr) 
	    + sizeof(PE_opthdr));

  if (l->l_info[DT_TEXTREL])
    {
      /* Bletch.  We must make read-only segments writable
	 long enough to relocate them.  */
      for(i=0; i<pFileHdr->f_nscns; i++, pSecHdr++) {
	if ((pSecHdr->s_flags & IMAGE_SCN_LNK_REMOVE) == 0
	    && (pSecHdr->s_flags & IMAGE_SCN_MEM_WRITE) == 0
	    && (pSecHdr->s_flags & IMAGE_SCN_MEM_DISCARDABLE) == 0) {

	    /* PE lets us assume page-aligned info here */
	    VMA mapstart = (VMA) l->l_addr + pSecHdr->s_vaddr;

	    if (_dl_sys_mprotect ((void *)mapstart, pSecHdr->s_size,
			    PROT_READ|PROT_WRITE) < 0)
	      _dl_signal_error (errno, l->l_name,
				"cannot make segment writable for relocation");
	  }
       }
    }

  {
    /* Do the actual relocation of the object's GOT and other data.  */

    const char *strtab		/* String table object symbols.  */
      = ((void *) l->l_addr + l->l_info[DT_STRTAB]);

    /* This macro is used as a callback from the DYNAMIC_RELOCATE code.  */
#define RESOLVE(ref, flags) \
    (_dl_lookup_symbol (pSymName, ref, scope, l->l_name, flags))

#include "dynamic-link.h"
    DYNAMIC_RELOCATE (l, lazy);
  }

  /* Set up the PLT so its unrelocated entries will jump to
     _dl_runtime_resolve (dl-runtime.c), which will relocate them.  */
  machine_runtime_setup (l, lazy);

  /* Mark the object so we know ths work has been done.  */
  l->l_relocated = 1;

  if (l->l_info[DT_TEXTREL])
    {
      /* Undo the protection change we made before relocating.  */
      pSecHdr = (PE_scnhdr *)(l->l_addr + sizeof(PE_filehdr) 
	    + sizeof(PE_opthdr));
      for(i=0; i<pFileHdr->f_nscns; i++, pSecHdr++) {
	if ((pSecHdr->s_flags & IMAGE_SCN_LNK_REMOVE) == 0
	    && (pSecHdr->s_flags & IMAGE_SCN_MEM_WRITE) == 0
	    && (pSecHdr->s_flags & IMAGE_SCN_MEM_DISCARDABLE) == 0) {
	    VMA mapstart = (VMA) l->l_addr + pSecHdr->s_vaddr;
	    int prot = 0;
	    if ((pSecHdr->s_flags & IMAGE_SCN_MEM_READ) != 0)
	      prot |= PROT_READ;
	    if ((pSecHdr->s_flags & IMAGE_SCN_MEM_EXECUTE) != 0)
	      prot |= PROT_EXEC;

	    /* PE lets us assume page-aligned info here */

	    if (_dl_sys_mprotect ((void *)mapstart, pSecHdr->s_size, prot) < 0)
	      _dl_signal_error (errno, l->l_name,
				"can't restore segment prot after reloc");
	  }
       }
    }
}
