/* Do relocations for dynamic linking.
   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/* Perform the relocations in MAP on the running program image as specified
   by RELTAG, SZTAG.  If LAZY is nonzero, this is the first pass on PLT
   relocations; they should be set up to call _dl_runtime_resolve, rather
   than fully resolved now.  */

static inline void
dynamic_do_rel (struct link_map *map,
		    int reltag, int sztag,
		    int lazy)
{
  VMA symtab = map->l_addr + map->l_info[DT_SYMTAB];
  VMA r = map->l_addr + map->l_info[reltag];
  VMA relend = r + map->l_info[sztag];
  PE_syment *symentry;

  /* slightly squrreley code because PE_RELSZ is 10, and subscripting
     doesn't work as expected (alignment); ditto for the symbol table */
  if (lazy)
    /* Doing lazy PLT relocations; they need very little info.  */
    for (; r<relend; r+=PE_RELSZ)
      machine_lazy_rel (map, (PE_reloc *)r);
  else
    for (; r<relend; r+=PE_RELSZ)
      {
         machine_rel (map, (PE_reloc *)r, symtab);
      }
}
