/* Inline functions for dynamic linking.
   Copyright (C) 1995, 1996, 1997 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <pe_headers.h>
#include <dl-machine.h>

/* Read the dynamic section at DYN and fill in INFO with indices DT_*.  */

static inline void __attribute__ ((unused))
get_dynamic_info (PE_dyn *dyn,
		      unsigned long info[DT_NUM + DT_PROCNUM + DT_EXTRANUM])
{
  int i;

  /* memset() won't inline (gcc bug?), so...; next line doesn't work either */
  /* for (i=0; i < ( sizeof(info)/sizeof(*info) ); i++) info[i]=0; */

  for (i=0; i < DT_NUM + DT_PROCNUM + DT_EXTRANUM; i++) info[i] = 0;

  if (! dyn)
    return;

  while (dyn->d_tag != DT_NULL)
    {
      if (dyn->d_tag < DT_NUM)
	info[dyn->d_tag] = dyn->d_un.d_val;
      else if (dyn->d_tag >= DT_LOPROC &&
	       dyn->d_tag < DT_LOPROC + DT_PROCNUM)
	info[dyn->d_tag - DT_LOPROC + DT_NUM] = dyn->d_un.d_val;
      else if ((int) DT_EXTRATAGIDX (dyn->d_tag) < DT_EXTRANUM)
	info[DT_EXTRATAGIDX (dyn->d_tag) + DT_NUM + DT_PROCNUM]=dyn->d_un.d_val;
      else
	_dl_assert (! "bad dynamic tag");

      /* TEXTREL is a present/absent boolean */
      if(dyn->d_tag == DT_TEXTREL)
	info[dyn->d_tag] = 1;

      /* DEBUG identifies a place for us to communicate back to the 
	 debugger; capture the address. */
      if(dyn->d_tag == DT_DEBUG)
	info[dyn->d_tag] = (VMA)&dyn->d_un.d_val;

      dyn++;
    }
}

#ifdef RESOLVE/* [ */

/* Get the definition of `dynamic_do_rel'
   It cannot be done in a more general function
   because we must be able to completely inline.  */

#include "do-rel.h"
#define DYNAMIC_DO_REL(map, lazy)					      \
  if ((map)->l_info[DT_REL])						      \
    dynamic_do_rel ((map), DT_REL, DT_RELSZ, 0);			      \
  if ((map)->l_info[DT_PLTREL] &&					      \
      (map)->l_info[DT_PLTREL] == DT_REL)			 	      \
    dynamic_do_rel ((map), DT_JMPREL, DT_PLTRELSZ, (lazy));

/* This can't just be an inline function because GCC is sometimes too dumb
   to inline functions containing inlines themselves.  */
#define DYNAMIC_RELOCATE(map, lazy) \
  do { DYNAMIC_DO_REL ((map), (lazy)); \
       } while (0)
#endif /* ] */
