/* Machine-dependent dynamic relocation inline functions.  i386 version.
   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#ifndef dl_machine_h
#define dl_machine_h

/*                      
 * Calculate how much off the link address we are loaded.
 * We want to keep relocateable expressions out of the text
 * segment, so go thru the GOT.
 */                             
#ifdef __PIC__
static inline VMA __attribute__ ((unused))
machine_load_offset (void)     
{ 
  /* This only works right BEFORE runtime relocation has occurred; the
     relocatons don't seem to provide a way to make a value absolute at
     link time.  Call before self-relocation and save the result. */
  VMA addr;
  asm ("        call 1f\n"
       "1:      popl  %0\n"
       "        mov xyzzy@GOT(%%ebx),%%eax\n"
       "        subl (%%eax),%0\n"
       "        .data\n"
       "xyzzy:	.long 1b\n"
       "	.text\n"
       : "=r" (addr) : : "eax", "cc");          
  return addr;
}                               
#else
#error NO non-PIC version of machine_load_offset()
#endif

/* Set up the loaded object described by L so its unrelocated PLT
   entries will jump to the on-demand fixup code in dl-runtime.c.  */

static inline void __attribute__ ((unused))
machine_runtime_setup (struct link_map *l, int lazy)
{
  VMA *got;
  extern void _dl_runtime_resolve (VMA);

  if (l->l_info[DT_JMPREL] && lazy)
    {
      /* The GOT entries for functions in the PLT have not yet been filled
	 in.  Their initial contents will arrange when called to push an
	 offset into the .rel.plt section, push _GLOBAL_OFFSET_TABLE_[1],
	 and then jump to _GLOBAL_OFFSET_TABLE[2].  */
      got = (VMA *) (l->l_addr + l->l_info[DT_PLTGOT]);
      got[1] = (VMA) l;	/* Identify this shared object.  */
      /* This function will get called to fix up the GOT entry indicated by
	 the offset on the stack, and then jump to the resolved address.  */
      got[2] = (VMA) &_dl_runtime_resolve;
    }

  /* This code is used in dl-runtime.c to call the `fixup' function
     and then redirect to the address it returns.  */
#define MACHINE_RUNTIME_TRAMPOLINE asm ("\
        .globl __dl_runtime_resolve
__dl_runtime_resolve:
	call _fixup	# Args pushed by PLT.
	addl $8, %esp	# Pop args.
	jmp *%eax	# Jump to function address.
");

#define machine_relplt machine_rel

}

#endif /* !dl_machine_h */

#ifdef RESOLVE

/* Perform the relocation specified by RELOC and SYM (which is fully resolved).
   MAP is the object containing the reloc.  */

extern struct link_map _dl_rtld_map;

static inline VMA
machine_rel (struct link_map *map,
		 const PE_reloc *reloc, VMA symtab)
{
  VMA *const reloc_addr = (void *) (map->l_addr + reloc->r_vaddr);
  VMA loadbase;
  const PE_syment *sym;
  int lookup_type;
  static VMA sym_value;
  static VMA lastSym;

  /* A switch would be very nice here (and much more readable), however
     we can't rely upon the compiler not to generate a jump table, and
     since we need this during bootstrap (when the jump table won't be
     relocated yet...) we can't do that. */
  /* This should be first in case it makes a difference in gen'd code */
  if (reloc->r_type == R_GNU_JUMP_SLOT) 
    {
      lookup_type = DL_LOOKUP_NOPLT;
    } 
#ifdef USE_COPY_RELOC
  else if (reloc->r_type == R_386_COPY)
    {
      lookup_type = DL_LOOKUP_NOEXEC;
    } 
#endif
  else if (reloc->r_type == R_PCRLONG
           || reloc->r_type == R_DIR32
           || reloc->r_type == R_GNU_GLOB_DAT)
    {
      lookup_type = 0;

    } 
  else if (reloc->r_type == R_GNU_RELATIVE)
    {
#ifndef RTLD_BOOTSTRAP
    if (map != &_dl_rtld_map) /* Already done in rtld itself.  */
#endif
      *reloc_addr += map->l_offs;
      return 0;
    } 

#if 0
  else if (reloc->r_type == 0)
    {
      /* This shouldn't happen, but in case of a need for fail-soft, code
	 is retained. */
      return 0;
    } 
#endif

  else 
    {
      _dl_assert (! "unexpected dynamic reloc type");
      return 0;
    }

  /* end virtual switch. */

  sym = (PE_syment *)(symtab + 
     ((VMA)(reloc->r_symndx)) * PE_SYMENT_SIZE);

  if ((VMA)sym+lookup_type != lastSym)
    {
      const char *pSymName;
      char tmpName[SYMNMLEN+1];

      /* We really are comparing two things: the symobl table entry address
	 and which of the 3 lookup types we are using; since symbol table
	 entries are > 3 bytes, this works nicely.  Excessivley cute, but
	 fast! */
      lastSym = (VMA)sym+lookup_type;

#ifndef RTLD_BOOTSTRAP
      if (sym->n_zeroes == 0) {
	 pSymName = strtab + sym->n_offset;
      }
      else {
	 *(PE_nameunion *)(tmpName) = sym->_n;
	 tmpName[SYMNMLEN] = 0;
	 pSymName = tmpName;
      }
#endif

      loadbase = RESOLVE (&sym, lookup_type);
      sym_value = sym ? (loadbase + sym->n_value) : 0;
    }

  switch (reloc->r_type)
    {
    /* This should be first in case it makes a difference in gen'd code */
    case R_GNU_JUMP_SLOT:
      *reloc_addr = sym_value;
      return sym_value;

#ifdef USE_COPY_RELOC
    case R_386_COPY:
      memcpy (reloc_addr, (void *) sym_avlue, sym->st_size);
      return 0;
#endif

    case R_GNU_GLOB_DAT:
      *reloc_addr = sym_value;
      return 0;

    case R_DIR32:
      {
	VMA undo = 0;
#ifndef RTLD_BOOTSTRAP
	if (map == &_dl_rtld_map)
	  /* Undo the relocation done here during bootstrapping.  Now we will
	     relocate it anew, possibly using a binding found in the user
	     program or a loaded library rather than the dynamic linker's
	     built-in definitions used while loading those libraries.  */
	  undo = map->l_addr + sym->n_value;
#endif
	*reloc_addr += sym_value - undo;
	return 0;
      }

    case R_PCRLONG:
      *reloc_addr += (sym_value - (VMA) reloc_addr) - 4;
      return 0;
    }
}

static inline void
machine_lazy_rel (struct link_map *map, const PE_reloc *reloc)
{
  VMA *const reloc_addr = (void *) (map->l_addr + reloc->r_vaddr);
  switch (reloc->r_type)
    {
    case R_GNU_JUMP_SLOT:
      *reloc_addr += map->l_offs;
      break;
    default:
      _dl_assert (! "unexpected PLT reloc type");
      break;
    }
}

#endif /* RESOLVE */
