/* PE Structures (suitable for non-bfd use).

   Copyright (C) 2000 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

/* This file reflects the structures of a PE executable (and .o) file.

   This is derived from a number of bfd and bfd-related headers with the
   following copyrights:
   .../bfd/../include/coff/pe.h: none provided.
   .../bfd/../include/coff/internal.h: none provided, 
                                       but asserted as part of bfd.
   .../bfd/../include/coff/i386.h: none provided.
   .../bfd/../include/elf/external.h: FSF 1991-1997 (GPL)
          (Dynamic data.)
   .../libexec/ld.so/ld.so/mips/elf.h: FSF 1996 (GLPL)
          (Dynamic data constants.)
*/

/* PE COFF header information */
#ifndef _PE_H
#define _PE_H

#include <sys/types.h>

/* COMDAT selection codes.  */

#define IMAGE_COMDAT_SELECT_NODUPLICATES     (1) /* Warn if duplicates.  */
#define IMAGE_COMDAT_SELECT_ANY		     (2) /* No warning.  */
#define IMAGE_COMDAT_SELECT_SAME_SIZE	     (3) /* Warn if different size.  */
#define IMAGE_COMDAT_SELECT_EXACT_MATCH	     (4) /* Warn if different.  */
#define IMAGE_COMDAT_SELECT_ASSOCIATIVE	     (5) /* Base on other section.  */

/* Magic values that are true for all dos/nt implementations */
#define DOSMAGIC       0x5a4d  
#define NT_SIGNATURE   0x00004550

#define IMAGE_FILE_MACHINE_I386  0x14c
#define IMAGE_FILE_MACHINE_ALPHA 0x184

#ifdef  _M_ALPHA
#define IMAGE_FILE_THIS_MACHINE IMAGE_FILE_MACHINE_ALPHA
#endif

#ifdef  _M_IX86
#define IMAGE_FILE_THIS_MACHINE IMAGE_FILE_MACHINE_I386
#endif

/* NT allows long filenames, we want to accommodate this. */
#define FILNMLEN	18	/* # characters in a file name		*/

typedef unsigned long VMA;

/********************** FILE HEADER **********************/
/* The "main" file header (at the beginning of an executable) */
typedef struct _PE_filehdr
{
  /* DOS header fields */
  ushort e_magic;		/* Magic number, 0x5a4d */
  ushort e_cblp;		/* Bytes on last page of file, 0x90 */
  ushort e_cp;			/* Pages in file, 0x3 */
  ushort e_crlc;		/* Relocations, 0x0 */
  ushort e_cparhdr;		/* Size of header in paragraphs, 0x4 */
  ushort e_minalloc;		/* Minimum extra paragraphs needed, 0x0 */
  ushort e_maxalloc;		/* Maximum extra paragraphs needed, 0xFFFF */
  ushort e_ss;			/* Initial (relative) SS value, 0x0 */
  ushort e_sp;			/* Initial SP value, 0xb8 */
  ushort e_csum;		/* Checksum, 0x0 */
  ushort e_ip;			/* Initial IP value, 0x0 */
  ushort e_cs;			/* Initial (relative) CS value, 0x0 */
  ushort e_lfarlc;		/* File address of relocation table, 0x40 */
  ushort e_ovno;		/* Overlay number, 0x0 */
  ushort e_res[4];		/* Reserved words, all 0x0 */
  ushort e_oemid;		/* OEM identifier (for e_oeminfo), 0x0 */
  ushort e_oeminfo;		/* OEM information; e_oemid specific, 0x0 */
  ushort e_res2[10];		/* Reserved words, all 0x0 */
  VMA    e_lfanew;		/* File address of new exe header, 0x80 */
  ulong  dos_message[16];	/* other stuff, always follow DOS header */
  ulong  nt_signature;		/* required NT signature, 0x4550 */ 

  /* From the coff standard header */  

  ushort f_magic;		/* magic number			*/
  ushort f_nscns;		/* number of sections		*/
  ulong  f_timdat;		/* time & date stamp		*/
  ulong  f_symptr;		/* file pointer to symtab	*/
  ulong  f_nsyms;		/* number of symtab entries	*/
  ushort f_opthdr;		/* sizeof(optional hdr)		*/
  ushort f_flags;		/* flags			*/

} PE_filehdr;

/* NT specific file attributes (in f_flags).  (BDF internal names also
   shown for consistency */
#define IMAGE_FILE_RELOCS_STRIPPED           0x0001
#define       F_RELFLG			     0x0001
#define IMAGE_FILE_EXECUTABLE_IMAGE          0x0002
#define       F_EXEC			     0x0001
#define IMAGE_FILE_LINE_NUMS_STRIPPED        0x0004
#define	      F_LNNO			     0x0004
#define IMAGE_FILE_LOCAL_SYMS_STRIPPED       0x0008
#define	      F_LSYMS			     0x0008
#define IMAGE_FILE_BYTES_REVERSED_LO         0x0080
#define IMAGE_FILE_32BIT_MACHINE             0x0100
#define IMAGE_FILE_DEBUG_STRIPPED            0x0200
#define	      F_DEBUG			     0x0200
#define IMAGE_FILE_SYSTEM                    0x1000
#define IMAGE_FILE_DLL                       0x2000
#define       F_DLL 			     0x2000
#define IMAGE_FILE_BYTES_REVERSED_HI         0x8000

#define FILHSZ 152

/********************** AOUT "OPTIONAL HEADER" **********************/

#define OMAGIC          0404    /* object files, eg as output */
#define ZMAGIC          0413    /* demand load format, eg normal ld output */
#define STMAGIC		0401	/* target shlib */
#define SHMAGIC		0443	/* host   shlib */

/* The "standard" part of the optional header */
typedef struct _aouthdr
{
  short	magic;			/* type of file				*/
  short	vstamp;			/* version stamp			*/
  long	tsize;			/* text size in bytes, padded to FW bdry*/
  long	dsize;			/* initialized data "  "		*/
  long	bsize;			/* uninitialized data "   "		*/
  VMA	entry;			/* entry pt.				*/
  VMA 	text_start;		/* base of text used for this file */
  VMA 	data_start;		/* base of data used for this file */
} AOUTHDR;

#define AOUTHDRSZ 28


/* extra structure which is used in the optional header */
typedef struct _IMAGE_DATA_DIRECTORY 
{
  long	VirtualAddress;
  long	Size;
} IMAGE_DATA_DIRECTORY;

#define IMAGE_NUMBEROF_DIRECTORY_ENTRIES  16

/* The NT additions to the COFF optional header */
typedef struct 
{
  AOUTHDR standard;

  VMA   ImageBase;			/* address of specific location in 
					   memory that file is located, 
					   NT default 0x10000 */
  long  SectionAlignment;		/* section alignment default 0x1000 */
  long  FileAlignment;			/* file alignment default 0x200 */
  short MajorOperatingSystemVersion;	/* minimum version of the operating */
  short MinorOperatingSystemVersion;	/* system req'd for exe, default to 1*/
  short MajorImageVersion;		/* user defineable field to store 
					   version of exe or dll being created*/
  short MinorImageVersion;		/* default to 0 */
  short MajorSubsystemVersion;		/* minimum subsystem version required */
  short MinorSubsystemVersion;		/* to run exe; default to 3.1 */
  long  Reserved1;			/* seems to be 0 */
  long  SizeOfImage;			/* size of mem to allocate for prog */
  long  SizeOfHeaders;			/* size of PE hdr and section table */
  long  CheckSum;			/* set to 0 */
  short Subsystem;
  /* type of subsystem exe uses for user interface,
     possible values:
     1 - NATIVE   Doesn't require a subsystem
     2 - WINDOWS_GUI runs in Windows GUI subsystem
     3 - WINDOWS_CUI runs in Windows char sub. (console app)
     5 - OS2_CUI runs in OS/2 character subsystem
     7 - POSIX_CUI runs in Posix character subsystem */
  short DllCharacteristics;		/* flags for DLL init, use 0 */
  long  SizeOfStackReserve;		/* amount of memory to reserve  */
  long  SizeOfStackCommit;		/* amount of memory initially committed
					   for initial thread's stack, default
					   is 0x1000 */
  long  SizeOfHeapReserve;		/* amount of virtual memory to reserve*/
  long  SizeOfHeapCommit;		/* and commit, don't know what to 
					   default it to */
  long  LoaderFlags;			/* can probably set to 0 */
  long  NumberOfRvaAndSizes;		/* number of entries in next entry, 16*/
  IMAGE_DATA_DIRECTORY DataDirectory[IMAGE_NUMBEROF_DIRECTORY_ENTRIES];
} PE_opthdr;

#define AOUTSZ (AOUTHDRSZ + 196)

/* define some NT default values */
#define NT_EXE_IMAGE_BASE    0x400000
#define NT_DLL_IMAGE_BASE    0x10000000
#define NT_SECTION_ALIGNMENT 0x1000
#define NT_FILE_ALIGNMENT    0x200
#define NT_DEF_RESERVE       0x100000
#define NT_DEF_COMMIT        0x1000
#define NT_DEF_SECTION_ALIGNMENT 0x1000
#define NT_DEF_FILE_ALIGNMENT 0x200


/********************** SECTION HEADER **********************/

#define SCNNMLEN (8)

typedef struct _PE_scnhdr {
    char   s_name[SCNNMLEN];	/* section name			*/
    VMA    s_paddr;		/* physical address, aliased s_nlib */
    VMA    s_vaddr;		/* virtual address		*/
    long   s_size;		/* section size			*/
    long   s_scnptr;		/* file ptr to raw data for section */
    long   s_relptr;		/* file ptr to relocation	*/
    ulong  s_lnnoptr;		/* file ptr to line numbers	*/
    ushort s_nreloc;		/* number of relocation entries	*/
    ushort s_nlnno;		/* number of line number entries*/
    ulong  s_flags;		/* flags			*/
} PE_scnhdr;

#define	SCNHSZ	40

/*
 * names of "special" sections
 */
#define _TEXT	".text"
#define _DATA	".data"
#define _BSS	".bss"

/*
 * Section characteristics
 */

#define IMAGE_SCN_TYPE_NO_PAD                0x00000008  /* Reserved. */
#define IMAGE_SCN_CNT_CODE                   0x00000020  /* code. */
#define IMAGE_SCN_CNT_INITIALIZED_DATA       0x00000040  /* init'd data. */
#define IMAGE_SCN_CNT_UNINITIALIZED_DATA     0x00000080  /* uninit'd data. */
#define IMAGE_SCN_LNK_OTHER                  0x00000100  /* Reserved. */
#define IMAGE_SCN_LNK_INFO                   0x00000200  /* comments or some */
						/* other type of information. */
#define IMAGE_SCN_LNK_REMOVE                 0x00000800  /* will not become */
						/* part of image. */
#define IMAGE_SCN_LNK_COMDAT                 0x00001000  /* comdat. */
#define IMAGE_SCN_MEM_FARDATA                0x00008000
#define IMAGE_SCN_MEM_PURGEABLE              0x00020000
#define IMAGE_SCN_MEM_16BIT                  0x00020000
#define IMAGE_SCN_MEM_LOCKED                 0x00040000
#define IMAGE_SCN_MEM_PRELOAD                0x00080000
#define IMAGE_SCN_ALIGN_1BYTES               0x00100000
#define IMAGE_SCN_ALIGN_2BYTES               0x00200000
#define IMAGE_SCN_ALIGN_4BYTES               0x00300000
#define IMAGE_SCN_ALIGN_8BYTES               0x00400000
#define IMAGE_SCN_ALIGN_16BYTES              0x00500000  /* Default alignment */
					       /* if no others are specified. */
#define IMAGE_SCN_ALIGN_32BYTES              0x00600000
#define IMAGE_SCN_ALIGN_64BYTES              0x00700000
#define IMAGE_SCN_LNK_NRELOC_OVFL            0x01000000  /* extended relocs. */
#define IMAGE_SCN_MEM_DISCARDABLE 	     0x02000000
#define IMAGE_SCN_MEM_NOT_CACHED             0x04000000  /* not cachable. */
#define IMAGE_SCN_MEM_NOT_PAGED              0x08000000  /* not pageable. */
#define IMAGE_SCN_MEM_SHARED                 0x10000000  /* shareable.    */
#define IMAGE_SCN_MEM_EXECUTE     	     0x20000000
#define IMAGE_SCN_MEM_READ        	     0x40000000
#define IMAGE_SCN_MEM_WRITE       	     0x80000000

/********************** SYMBOLS **********************/

/* Relocatable symbols have number of the section in which they are defined,
   or one of the following: */

#define N_UNDEF	((short)0)	/* undefined symbol */
#define N_ABS	((short)-1)	/* value of symbol is absolute */
#define N_DEBUG	((short)-2)	/* debugging symbol -- value is meaningless */

/*
 * Type of a symbol, in low N bits of the word
 */
#define T_NULL		0
#define T_VOID		1	/* function argument (only used by compiler) */
#define T_CHAR		2	/* character		*/
#define T_SHORT		3	/* short integer	*/
#define T_INT		4	/* integer		*/
#define T_LONG		5	/* long integer		*/
#define T_FLOAT		6	/* floating point	*/
#define T_DOUBLE	7	/* double word		*/
#define T_STRUCT	8	/* structure 		*/
#define T_UNION		9	/* union 		*/
#define T_ENUM		10	/* enumeration 		*/
#define T_MOE		11	/* member of enumeration*/
#define T_UCHAR		12	/* unsigned character	*/
#define T_USHORT	13	/* unsigned short	*/
#define T_UINT		14	/* unsigned integer	*/
#define T_ULONG		15	/* unsigned long	*/
#define T_LNGDBL	16	/* long double		*/

/*
 * derived types, in n_type
*/
#define DT_NON		(0)	/* no derived type */
#define DT_PTR		(1)	/* pointer */
#define DT_FCN		(2)	/* function */
#define DT_ARY		(3)	/* array */

#define N_BTMASK	(0xf)
#define N_TMASK		(0x30)
#define N_BTSHFT	(4)
#define N_TSHIFT	(2)

#define BTYPE(x)	((x) & N_BTMASK)
#define DTYPE(x)        (((x) & N_TMASK) >>  N_BTSHFT)

#define ISPTR(x)	(((x) & N_TMASK) == (DT_PTR << N_BTSHFT))
#define ISFCN(x)	(((x) & N_TMASK) == (DT_FCN << N_BTSHFT))
#define ISARY(x)	(((x) & N_TMASK) == (DT_ARY << N_BTSHFT))
#define ISTAG(x)	((x)==C_STRTAG||(x)==C_UNTAG||(x)==C_ENTAG)
#define DECREF(x) ((((x)>>N_TSHIFT)&~N_BTMASK)|((x)&N_BTMASK))

/*
 * Storage Class of a symbol
 */

/* This used to be defined as -1, but now n_sclass is unsigned.  */
#define C_EFCN		0xff	/* physical end of function	*/
#define C_NULL		0
#define C_AUTO		1	/* automatic variable		*/
#define C_EXT		2	/* external symbol		*/
#define C_STAT		3	/* static			*/
#define C_REG		4	/* register variable		*/
#define C_EXTDEF	5	/* external definition		*/
#define C_LABEL		6	/* label			*/
#define C_ULABEL	7	/* undefined label		*/
#define C_MOS		8	/* member of structure		*/
#define C_ARG		9	/* function argument		*/
#define C_STRTAG	10	/* structure tag		*/
#define C_MOU		11	/* member of union		*/
#define C_UNTAG		12	/* union tag			*/
#define C_TPDEF		13	/* type definition		*/
#define C_USTATIC	14	/* undefined static		*/
#define C_ENTAG		15	/* enumeration tag		*/
#define C_MOE		16	/* member of enumeration	*/
#define C_REGPARM	17	/* register parameter		*/
#define C_FIELD		18	/* bit field			*/
#define C_AUTOARG	19	/* auto argument		*/
#define C_LASTENT	20	/* dummy entry (end of block)	*/
#define C_BLOCK		100	/* ".bb" or ".eb"		*/
#define C_FCN		101	/* ".bf" or ".ef"		*/
#define C_EOS		102	/* end of structure		*/
#define C_FILE		103	/* file name			*/
#define C_LINE		104	/* line # reformatted as symbol table entry */
#define C_ALIAS	 	105	/* duplicate tag		*/
#define C_HIDDEN	106	/* ext symbol in dmert public lib */
#define C_SECTION       104     /* section name */
#define C_NT_WEAK	105	/* weak external */

/*
 * A symbol entry
 */

#define SYMNMLEN	8	/* # characters in a symbol name	*/

/* NOTE:
   The PE_syment structure is 18 (yes, 18) bytes long.  On many machines,
   this will cause problems because it isn't properly (4-byte) aligned.
   On NT, there is a possible performance concern because on machines 
   like the Alpha, this will appear to work, but there will be alignment
   traps and fixups */

typedef union _PE_nameunion
{
  char n_name[SYMNMLEN];
  struct {
    ulong _n_zeroes;
    ulong _n_offset;
  } _n_n;
} PE_nameunion;

typedef struct _PE_syment 
{
  PE_nameunion _n;
  long   n_value;
  short  n_scnum;
  ushort n_type;
  unsigned char  n_sclass;
  unsigned char  n_numaux;
} PE_syment;

#define n_name		_n._n_name
#define n_zeroes	_n._n_n._n_zeroes
#define n_offset	_n._n_n._n_offset

#define PE_SYMENT_SIZE 18

/********************** SYMBOL TABLE AUX ENTRIES **********************/

/* By definition, these are the same size as symbol table entries, and
   thus the comment above applies here, too. */

#define DIMNUM		4	/* # array dimensions in auxiliary entry */

typedef union _PE_auxent {
    struct {
	long x_tagndx;			/* str, un, or enum tag indx */
	union {
	    struct {
		unsigned short x_lnno;	/* declaration line number */
		unsigned short x_size;	/* str/union/array size */
	    } x_lnsz;
	    long x_fsize;		/* size of function */
	} x_misc;

	union {
	    struct {			/* if ISFCN, tag, or .bb */
		long x_lnnoptr;		/* ptr to fcn line # */
		long x_endndx;		/* entry ndx past block end */
	    } x_fcn;
	    struct {			/* if ISARY, up to 4 dimen. */
		unsigned short x_dimen[DIMNUM];
	    } x_ary;
	} x_fcnary;
	unsigned short x_tvndx;		/* tv index */
    } x_sym;

    union {
	char x_fname[FILNMLEN];
	struct {
	    long x_zeroes;
	    long x_offset;
	} x_n;
    } x_file;

    struct {
	long x_scnlen;	        	/* section length */
	unsigned short x_nreloc;	/* # relocation entries */
	unsigned short x_nlinno;	/* # line numbers */
	unsigned long x_checksum;	/* section COMDAT checksum */
	unsigned short x_associated;	/* COMDAT associated section index */
	unsigned char x_comdat;		/* COMDAT selection number */
    } x_scn;

    struct {
	long x_tvfill[4];		/* tv fill value */
	unsigned short x_tvlen;		/* length of .tv */
	unsigned short x_tvran[2];	/* tv range */
    } x_tv;		/* info about .tv section (in auxent of symbol .tv)) */
} PE_auxent;

#define	PE_AUXENT_SIZE PE_SYMENT_SIZE

/********************** RELOCATION DIRECTIVES **********************/

/* See note above... another odd size */
typedef struct _PE_reloc {
  VMA    r_vaddr;
  long   r_symndx;
  ushort r_type;
} PE_reloc;

#define PE_RELSZ 10

#define	R_DIR16		01
#define R_DIR32		06
#define R_IMAGEBASE     07

#define R_RELBYTE	017
#define R_RELWORD	020
#define	R_PCLONG	020
#define R_RELLONG	021
#define R_PCRBYTE	022
#define R_PCRWORD	023
#define R_PCRLONG	024

/* These must be in the same order as the corr. BFD_RELOC_386... symbols */
#define R_GNU_GOT32     030
#define R_GNU_PLT32     031
#define R_GNU_COPY      032
#define R_GNU_GLOB_DAT  033
#define R_GNU_JUMP_SLOT 034
#define R_GNU_RELATIVE  035
#define R_GNU_GOTOFF    036
#define R_GNU_GOTPC     037

/* The r_type field in a reloc is one of the following values.  */
#define ALPHA_R_IGNORE		0
#define ALPHA_R_REFLONG		1
#define ALPHA_R_REFQUAD		2
#define ALPHA_R_GPREL32		3
#define ALPHA_R_LITERAL		4
#define ALPHA_R_LITUSE		5
#define ALPHA_R_GPDISP		6
#define ALPHA_R_BRADDR		7
#define ALPHA_R_HINT		8

/* Specific to NT */
#define IMAGE_REL_ALPHA_INLINE_REFLONG  9
#define IMAGE_REL_ALPHA_REFHI          10
#define IMAGE_REL_ALPHA_REFLO          11
#define IMAGE_REL_ALPHA_PAIR           12
#define IMAGE_REL_ALPHA_MATCH          13
#define IMAGE_REL_ALPHA_SECTION        14
#define IMAGE_REL_ALPHA_SECREL         15
#define IMAGE_REL_ALPHA_REFLONGNB      16

/* In the NT model we need some extra runtime relocations; we use the
   ones above when they apply. */
#define ALPHA_DYN_R_REL_32		101
#define ALPHA_DYN_R_REL_64		102
#define ALPHA_DYN_R_JMP_SLOT		103
#define ALPHA_DYN_R_REL_REFHI		104
#define ALPHA_DYN_R_REL_REFLO		105
#define ALPHA_DYN_R_REL_PAIR		106
#define ALPHA_DYN_R_COMB		107
#define ALPHA_DYN_R_REL_HIONLY		108

/********************** DYNAMIC SECTION **********************/

typedef struct {
  ulong	d_tag;		/* entry tag value */
  union {
    ulong	d_val;
    VMA		d_ptr;
  } d_un;
} PE_dyn;

/* Legal values for d_tag (dynamic entry type).  */

#define DT_NULL		0		/* Marks end of dynamic section */
#define DT_NEEDED	1		/* Name of needed library */
#define DT_PLTRELSZ	2		/* Size in bytes of PLT relocs */
#define DT_PLTGOT	3		/* Processor defined value */
#define DT_HASH		4		/* Address of symbol hash table */
#define DT_STRTAB	5		/* Address of string table */
#define DT_SYMTAB	6		/* Address of symbol table */
#define DT_RELA		7		/* Address of Rela relocs */
#define DT_RELASZ	8		/* Total size of Rela relocs */
#define DT_RELAENT	9		/* Size of one Rela reloc */
#define DT_STRSZ	10		/* Size of string table */
#define DT_SYMENT	11		/* Size of one symbol table entry */
#define DT_INIT		12		/* Address of init function */
#define DT_FINI		13		/* Address of termination function */
#define DT_SONAME	14		/* Name of shared object */
#define DT_RPATH	15		/* Library search path */
#define DT_SYMBOLIC	16		/* Start symbol search here */
#define DT_REL		17		/* Address of Rel relocs */
#define DT_RELSZ	18		/* Total size of Rel relocs */
#define DT_RELENT	19		/* Size of one Rel reloc */
#define DT_PLTREL	20		/* Type of reloc in PLT */
#define DT_DEBUG	21		/* For debugging; unspecified */
#define DT_TEXTREL	22		/* Reloc might modify .text */
#define DT_JMPREL	23		/* Address of PLT relocs */
#define	DT_NUM		24		/* Number used.  */
#define DT_LOPROC	0x70000000	/* Start of processor-specific */
#define DT_HIPROC	0x7fffffff	/* End of processor-specific */
#define DT_MAXPROCUSED  0x70000000      /* Last processor-specific act. used */

/* The next four dynamic tags are used on Solaris.  We support them
   everywhere.  */
#define DT_VERDEF       0x6ffffffc
#define DT_VERDEFNUM    0x6ffffffd
#define DT_VERNEED      0x6ffffffe
#define DT_VERNEEDNUM   0x6fffffff

/* These section tags are used on Solaris.  We support them
   everywhere, and hope they do not conflict.  */

#define DT_AUXILIARY    0x7ffffffd
#define DT_FILTER       0x7fffffff
#define DT_EXTRATAGIDX(tag)	((int) -((long)(tag) <<1>>1) - 1)
#define DT_EXTRANUM	3

/* This tag is a GNU extension to the Solaris version scheme.  */
#define DT_VERSYM       0x6ffffff0

/********************** LINE NUMBERS **********************/

/* 1 line number entry for every "breakpointable" source line in a section.
 * Line numbers are grouped on a per function basis; first entry in a function
 * grouping will have l_lnno = 0 and in place of physical address will be the
 * symbol table index of the function name.
 */
typedef struct _PE_lineno {
    union {
	long l_symndx;	/* function name symbol index, iff l_lnno == 0 */
	long l_paddr;	/* (physical) address of line number	*/
    } l_addr;
    unsigned short l_lnno;	/* line number */
} PE_lineno;

#define	LINESZ	6

#endif /* _PE_H */
