/* Run time dynamic linker.
   Copyright (C) 1995, 1996, 1997 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <unistd.h>

#include <link.h>
#include "dl-internal.h"
#include "dynamic-link.h"
#include "syscall.h"
#include <interix/ldsostartup.h>

extern char **_dl_environ;
extern char **_dl_argv;
extern int _dl_argc;

const char* _dl_rpath;

struct link_map _dl_rtld_map;  /* the map entry for ld.so */

int
_dl_start (int startmode, startupData *startupData, void *startaddr)
{
  struct link_map bootstrap_map;

#ifdef RELOCATE_BOOTSTRAP_LOADER

  /* This #define produces dynamic linking inline functions for
     bootstrap relocation instead of general-purpose relocation.  */
#define RTLD_BOOTSTRAP
#define RESOLVE(sym, flags) bootstrap_map.l_addr
#include "dynamic-link.h"

#endif

  /* Figure out the run-time load address of the dynamic linker itself.  */
  bootstrap_map.l_addr = startupData->ldsoLoadAddr;
  bootstrap_map.l_offs = machine_load_offset();

#ifndef RELOCATE_BOOTSTRAP_LOADER
  if (bootstrap_map.l_offs != 0)
    {
	no_relocate_message(bootstrap_map.l_offs);
	/* This is probably noreturn */
    }
#endif

  /* Read our own dynamic section and fill in the info array.  */
  bootstrap_map.l_ld = (void *) bootstrap_map.l_offs + (VMA)&_DYNAMIC;
  get_dynamic_info (bootstrap_map.l_ld, bootstrap_map.l_info);

  /* Relocate ourselves so we can do normal function calls and
     data access using the global offset table.  */

#ifdef RELOCATE_BOOTSTRAP_LOADER
  DYNAMIC_RELOCATE (&bootstrap_map, 0);
#endif

  /* Now life is sane; we can call functions and access global data.
     Set up to use the operating system facilities, and find out from
     the operating system's program loader where to find the program
     header table in core.  */

  /* Set the address we use for errno; we always access errno thru a pointer,
     as the address is different when we're bootstrapping thru ld.so
     than when we're called after main() has been called.  */
  _dl_perrno = startupData->pErrno;    /* the temporary value */
  startupData->ppErrno = &_dl_perrno;  /* where to put the permanent value */
  /* (If we're lucky, the temprary value IS the permanent value, but...) */

  /* Transfer data about ourselves to the permanent link_map structure.  */
  _dl_rtld_map.l_addr = bootstrap_map.l_addr;
  _dl_rtld_map.l_ld = bootstrap_map.l_ld;
  memcpy (_dl_rtld_map.l_info, bootstrap_map.l_info,
	  sizeof _dl_rtld_map.l_info);
  _dl_setup_hash (&_dl_rtld_map);

  /* Cache the DT_RPATH stored in ld.so itself; this will be
     the default search path.  */
  if (_dl_rtld_map.l_info[DT_RPATH] != 0)
      _dl_rpath = (void *) (_dl_rtld_map.l_addr +
			_dl_rtld_map.l_info[DT_STRTAB] +
			_dl_rtld_map.l_info[DT_RPATH]);

  /* Tell everyone else who got loaded */
  _dl_rtld_map.l_libname = startupData->ldsoName;

  _dl_environ = startupData->environ;
  _dl_argc = startupData->argc;
  _dl_argv = startupData->argv;

  /* Call the OS-dependent function to set up life so we can do things like
     file access.  It will call `dl_main' (below) to do all the real work
     of the dynamic linker, and then unwind our frame and run the user
     entry point on the same stack we entered on.  */
  _dl_main(startupData->progLoadAddr, startmode, startaddr);

  /* _dl_main can change these, so put 'em back right */
  startupData->argc = _dl_argc;
  startupData->argv = _dl_argv;

  startupData->fini = _dl_fini;
}

void _dl_static_init(void)
{
  /* do nothing... the static case wants this. */
}
